# BSD 3 - Clause License

# Copyright(c) 2021, Zenotech
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and / or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#         SERVICES
#         LOSS OF USE, DATA, OR PROFITS
#         OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
import importlib
import logging

from subprocess import check_output, CalledProcessError

from .exceptions import ConfigurationException

logger = logging.getLogger(__name__)

_available_schedulers = {
    "slurm": "mycluster.schedulers.slurm.Slurm",
    "pbs": "mycluster.schedulers.pbs.PBS",
    "lsf": "mycluster.schedulers.lsf.LSF",
    "sge": "mycluster.schedulers.sge.SGE",
}


def my_import(name):
    logger.debug(f"Importing schduler class {name}")
    module = mod = name.rsplit(".", 1)[0]
    klass = mod = name.rsplit(".", 1)[1]
    mod = importlib.import_module(module)
    return getattr(mod, klass)


def get_scheduler(scheduler_name):
    logger.debug(f"Getting implementation of {scheduler_name}")
    if scheduler_name in _available_schedulers:
        return my_import(_available_schedulers[scheduler_name])()
    else:
        raise ConfigurationException(f"Invalid scheduler specified: {scheduler_name}")


def detect_scheduler_name():
    """
    Try to automatically detect scheduler type
    """
    # Test for custom scheduler
    if os.getenv("MYCLUSTER_SCHED") is not None:
        logger.debug(f"Custom scheduler defined using MYCLUSTER_SCHED")
        return os.getenv("MYCLUSTER_SCHED")

    # Test for SLURM
    if os.getenv("SLURMHOME") is not None:
        logger.debug(f"SLURMHOME set, Slurm assumed")
        return "slurm"

    try:
        line = check_output(["scontrol", "ping"], universal_newlines=True)
        if line.split("(")[0] == "Slurmctld":
            logger.debug(f"scontrol responded, Slurm assumed")
            return "slurm"
    except (CalledProcessError, FileNotFoundError) as e:
        pass

    # Test for PBS
    try:
        line = check_output(["pbsnodes", "-a"])
        logger.debug(f"pbsnodes responded, PBS assumed")
        return "pbs"
    except (CalledProcessError, FileNotFoundError) as e:
        pass

    # Test for SGE
    if os.getenv("SGE_CLUSTER_NAME") is not None:
        logger.debug(f"SGE_CLUSTER_NAME set, SGE assumed")
        return "sge"

    # Test for lsf
    try:
        line = check_output("lsid", universal_newlines=True)
        if line.split(" ")[0] == "Platform" or line.split(" ")[0] == "IBM":
            logger.debug(f"lsid responded, LSF assumed")
            return "lsf"
    except (CalledProcessError, FileNotFoundError) as e:
        pass

    logger.debug(f"All detection tests failed, unabled to detect local scheduler")
    raise ConfigurationException("Error, unabled to detect local scheduler")


def detect_scheduling_sys():
    """
    Returns the scheduler implementation
    """
    name = detect_scheduler_name()
    return get_scheduler(name)



# BSD 3 - Clause License

# Copyright(c) 2021, Zenotech
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and / or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#         SERVICES
#         LOSS OF USE, DATA, OR PROFITS
#         OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
import re
import math
import subprocess
import datetime

from .base import Scheduler
from mycluster.exceptions import SchedulerException, ConfigurationException


class SGE(Scheduler):
    def scheduler_type(self):
        return "sge"

    def name(self):
        return os.getenv("SGE_CLUSTER_NAME")

    def queues(self):
        # list all parallel env
        # for parallel_env list queues associated
        # Find first node with queue and record node config
        queue_list = []
        parallel_env_list = []
        with os.popen("qconf -spl") as f:
            for line in f:
                parallel_env_list.append(line.strip())
        for parallel_env in parallel_env_list:
            with os.popen("qstat -pe " + parallel_env + " -U `whoami` -g c") as f:
                f.readline()  # read header
                f.readline()  # read separator
                for line in f:
                    queue_name = line.split(" ")[0].strip()
                    # Check if user has permission to use queue
                    with os.popen("qstat -g c -U `whoami` -q " + queue_name) as f2:
                        try:
                            f2.readline()
                            f2.readline()
                            if len(f2.readline()):
                                queue_list.append(parallel_env + ":" + queue_name)
                        except:
                            pass
        return queue_list

    def node_config(self, queue_id):
        # Find first node with queue and record node config
        parallel_env = queue_id.split(":")[0]
        queue_name = queue_id.split(":")[1]
        host_group = 0
        with os.popen("qconf -sq " + queue_name) as f:
            for line in f:
                if line.split(" ")[0] == "hostlist":
                    new_line = re.sub(" +", " ", line)
                    host_group = new_line.split(" ")[1]

        config = {}
        host_name = ""
        found = False
        if host_group[0] is "@":
            # Is a host group
            with os.popen("qconf -shgrp_resolved " + host_group) as f:
                for line in f:
                    for host_name in line.split(" "):
                        with os.popen("qhost -q -h " + host_name) as f:
                            header = f.readline()  # read header
                            f.readline()  # read separator
                            new_header = re.sub(" +", " ", header).strip()
                            # sge <=6.2u4 style
                            if (new_header.split(" ")[3]) == "LOAD":
                                for line in f:
                                    if line[0] != " ":
                                        name = line.split(" ")[0]
                                        if name != "global":
                                            new_line = re.sub(" +", " ", line).strip()
                                            if new_line.split(" ")[3] != "-":
                                                config["max task"] = int(
                                                    new_line.split(" ")[2]
                                                )
                                                config["max thread"] = int(
                                                    new_line.split(" ")[2]
                                                )
                                                config["max memory"] = new_line.split(
                                                    " "
                                                )[4]
                                                found = True
                                                break
                            else:
                                for line in f:
                                    if line[0] != " ":
                                        name = line.split(" ")[0]
                                        if name != "global":
                                            new_line = re.sub(" +", " ", line).strip()
                                            if new_line.split(" ")[3] != "-":
                                                config["max task"] = int(
                                                    new_line.split(" ")[4]
                                                )
                                                config["max thread"] = int(
                                                    new_line.split(" ")[5]
                                                )
                                                config["max memory"] = new_line.split(
                                                    " "
                                                )[7]
                                                found = True
                                                break
                        if found:
                            break
        else:
            # Is a host
            host_name = host_group
            with os.popen("qhost -q -h " + host_name) as f:
                header = f.readline()  # read header
                f.readline()  # read separator
                new_header = re.sub(" +", " ", header).strip()
                if (new_header.split(" ")[3]) == "LOAD":  # sge <=6.2u4 style
                    for line in f:
                        if line[0] != " ":
                            name = line.split(" ")[0]
                            if name != "global":
                                new_line = re.sub(" +", " ", line).strip()
                                if new_line.split(" ")[3] != "-":
                                    config["max task"] = int(new_line.split(" ")[2])
                                    config["max thread"] = int(new_line.split(" ")[2])
                                    config["max memory"] = new_line.split(" ")[4]
                                else:
                                    config["max task"] = 0
                                    config["max thread"] = 0
                                    config["max memory"] = 0
                else:
                    for line in f:
                        if line[0] != " ":
                            name = line.split(" ")[0]
                            if name != "global":
                                new_line = re.sub(" +", " ", line).strip()
                                if new_line.split(" ")[3] != "-":
                                    config["max task"] = int(new_line.split(" ")[4])
                                    config["max thread"] = int(new_line.split(" ")[5])
                                    config["max memory"] = new_line.split(" ")[7]
                                else:
                                    config["max task"] = 0
                                    config["max thread"] = 0
                                    config["max memory"] = 0
        return config

    def tasks_per_node(self, queue_id):
        parallel_env = queue_id.split(":")[0]
        queue_name = queue_id.split(":")[1]
        tasks = 0
        with os.popen("qconf -sq " + queue_name) as f:
            for line in f:
                if line.split(" ")[0] == "slots":
                    tasks = int(re.split("\W+", line)[1])

        pe_tasks = tasks
        with os.popen("qconf -sp " + parallel_env) as f:
            try:
                for line in f:
                    if line.split(" ")[0] == "allocation_rule":
                        try:
                            # This may throw exception as allocation rule
                            # may not always be an integer
                            pe_tasks = int(re.split("\W+", line)[1])
                        except ValueError as e:
                            raise SchedulerException("Error parsing SGE output")
            except:
                pass

        return min(tasks, pe_tasks)

    def available_tasks(self, queue_id):
        # split queue id into queue and parallel env
        # list free slots
        free_tasks = 0
        max_tasks = 0
        parallel_env = queue_id.split(":")[0]
        queue_name = queue_id.split(":")[1]
        with os.popen(" qstat -pe " + parallel_env + " -U `whoami` -g c") as f:
            f.readline()  # read header
            f.readline()  # read separator
            for line in f:
                # remove multiple white space
                new_line = re.sub(" +", " ", line)
                qn = new_line.split(" ")[0]
                if qn == queue_name:
                    free_tasks = int(new_line.split(" ")[4])
                    max_tasks = int(new_line.split(" ")[5])

        return {"available": free_tasks, "max tasks": max_tasks}

    def _min_tasks_per_node(self, queue_id):
        """
        This function is used when requesting non exclusive use
        as the parallel environment might enforce a minimum number
        of tasks
        """
        parallel_env = queue_id.split(":")[0]
        queue_name = queue_id.split(":")[1]
        tasks = 1
        pe_tasks = tasks
        with os.popen("qconf -sp " + parallel_env) as f:
            try:
                for line in f:
                    if line.split(" ")[0] == "allocation_rule":
                        # This may throw exception as allocation rule
                        # may not always be an integer
                        pe_tasks = int(re.split("\W+", line)[1])
            except:
                pass

        return max(tasks, pe_tasks)

    def create_submit(
        self,
        queue_id,
        num_tasks,
        job_name,
        job_script,
        wall_clock,
        openmpi_args="-bysocket -bind-to-socket",
        project_name="default",
        tasks_per_node=None,
        threads_per_task=1,
        user_email=None,
        qos=None,
        exclusive=True,
        output_name=None,
        gres=None
    ):
        parallel_env = queue_id.split(":")[0]
        queue_name = queue_id.split(":")[1]

        if tasks_per_node is None:
            tasks_per_node = self.tasks_per_node(queue_id)

        num_nodes = int(math.ceil(float(num_tasks) / float(tasks_per_node)))
        if threads_per_task is None:
            threads_per_task = 1

        if ":" not in wall_clock:
            wall_clock = wall_clock + ":00:00"

        if "mycluster-" in job_script:
            job_script = self._get_data(job_script)

        if output_name is None:
            output_name = job_name + ".out"

        # For exclusive node use total number of slots required
        # is number of nodes x number of slots offer by queue
        num_queue_slots = num_nodes * self.tasks_per_node(queue_id)

        if not exclusive:
            if num_nodes == 1:  # Assumes fill up rule
                num_queue_slots = max(
                    tasks_per_node, self._min_tasks_per_node(queue_id)
                )

        template = self._load_template("sge.jinja")

        script_str = template.render(
            my_name=job_name,
            my_script=job_script,
            my_output=output_name,
            user_email=user_email,
            queue_name=queue_name,
            parallel_env=parallel_env,
            num_queue_slots=num_queue_slots,
            num_tasks=num_tasks,
            tpn=tasks_per_node,
            num_threads_per_task=threads_per_task,
            num_nodes=num_nodes,
            project_name=project_name,
            wall_clock=wall_clock,
            openmpi_args=openmpi_args,
            qos=qos,
            exclusive=exclusive,
        )

        return script_str

    def submit(
        self, script_name, immediate=False, depends_on=None, depends_on_always_run=False
    ):
        job_id = None

        output = subprocess.run(
            f"qsub -V -terse {script_name}",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
        )
        if output.returncode == 0:
            job_id = 0
            out = output.stdout.decode("utf-8")
            try:
                job_id = int(out.readline().strip())
                return job_id
            except:
                raise SchedulerException("Error submitting job to SGE")
        else:
            raise SchedulerException(
                f"Error submitting job to SGE: {output.stderr.decode('utf-8')}"
            )

    def list_current_jobs(self):
        jobs = []
        output = subprocess.run(
            "qstat -u `whoami`",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
        )
        if output.returncode == 0:
            for line in output.stdout.decode("utf-8").splitlines():
                if line.startswith("job-ID") or line.startswith("---"):
                    continue
                job_info = re.sub(" +", " ", line.strip()).split(" ")
                jobs.append(
                    {
                        "id": int(job_info[0]),
                        "queue": job_info[6],
                        "name": job_info[2],
                        "state": job_info[4],
                    }
                )
        else:
            raise SchedulerException("Error fetching job queue listing")
        return jobs

    def get_job_details(self, job_id):
        """
        Get full job and step stats for job_id
        First check using sacct, then fallback to squeue
        """
        stats_dict = {}
        output = {}
        with os.popen("qacct -j " + str(job_id)) as f:
            try:
                f.readline()  # read header
                for line in f:
                    new_line = re.sub(" +", " ", line.strip())
                    output[new_line.split(" ")[0]] = new_line.split(" ", 1)[1]
            except:
                pass
        stats_dict["wallclock"] = datetime.timedelta(
            seconds=int(output["ru_wallclock"])
        )
        stats_dict["mem"] = output["mem"]
        stats_dict["cpu"] = datetime.timedelta(seconds=int(output["cpu"].split(".")[0]))
        stats_dict["queue"] = output["granted_pe"] + ":" + output["qname"]
        return stats_dict

    def delete(self, job_id):
        cmd = f"qdel {job_id}"
        output = subprocess.run(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
        )
        if output.returncode != 0:
            raise SchedulerException(f"Error cancelling job {job_id}")



# BSD 3 - Clause License

# Copyright(c) 2021, Zenotech
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and / or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#         SERVICES
#         LOSS OF USE, DATA, OR PROFITS
#         OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
import re
import math
import subprocess
import datetime

from .base import Scheduler
from mycluster.exceptions import SchedulerException, ConfigurationException


class LSF(Scheduler):
    def scheduler_type(self):
        return "lsf"

    def name(self):
        lsid_output = self._check_output(["lsid"]).splitlines()
        for line in lsid_output:
            if line.startswith("My cluster name is"):
                return line.rsplit(" ", 1)[1].strip()
        return "undefined"

    def queues(self):
        queue_list = []
        with os.popen("bqueues -w -u `whoami`") as f:
            f.readline()  # read header
            for line in f:
                q = line.split(" ")[0].strip()
                queue_list.append(q)
        return queue_list

    def node_config(self, queue_id):
        # Find first node with queue and record node config
        # bqueues -l queue_id
        host_list = None
        config = {}
        q_output = self._check_output(["bqueues", "-l", queue_id]).splitlines()
        for line in q_output:
            if line.startswith("HOSTS:"):
                host_list = line.strip().rsplit(" ", 1)[1].replace("/", "")
                if host_list == "none":
                    config["max task"] = 0
                    config["max thread"] = 0
                    config["max memory"] = 0
                    return config
        bhosts_output = self._check_output(["bhosts", "-l", host_list]).splitlines()
        line = re.sub(" +", " ", bhosts_output[2]).strip()
        tasks = int(line.split(" ")[3])
        line = re.sub(" +", " ", bhosts_output[6]).strip()
        memory = float(line.split(" ")[11].replace("G", ""))
        config["max task"] = tasks
        config["max thread"] = tasks
        config["max memory"] = memory
        return config

    def tasks_per_node(self, queue_id):
        host_list = None
        q_output = self._check_output(["bqueues", "-l", queue_id]).splitlines()
        for line in q_output:
            if line.startswith("HOSTS:"):
                host_list = line.strip().rsplit(" ", 1)[1].replace("/", "")
                if host_list == "none":
                    return 1
        bhosts_output = self._check_output(["bhosts", "-l", host_list]).splitlines()
        line = re.sub(" +", " ", bhosts_output[2]).strip()
        tasks = int(line.split(" ")[3])
        return tasks

    def available_tasks(self, queue_id):
        # split queue id into queue and parallel env
        # list free slots
        free_tasks = 0
        max_tasks = 0
        run_tasks = 0
        queue_name = queue_id
        q_output = self._check_output(["bqueues", queue_name]).splitlines()
        for line in q_output:
            if line.startswith(queue_name):
                new_line = re.sub(" +", " ", line).strip()
                try:
                    max_tasks = int(new_line.split(" ")[4])
                except:
                    pass
                pen_tasks = int(new_line.split(" ")[8])
                run_tasks = int(new_line.split(" ")[9])
                sus_tasks = int(new_line.split(" ")[10])
        return {"available": max_tasks - run_tasks, "max tasks": max_tasks}

    def create_submit(
        self,
        queue_id,
        num_tasks,
        job_name,
        job_script,
        wall_clock,
        openmpi_args="-bysocket -bind-to-socket",
        project_name="default",
        tasks_per_node=None,
        threads_per_task=1,
        user_email=None,
        qos=None,
        exclusive=True,
        output_name=None,
        gres=None
    ):
        if tasks_per_node is None:
            tasks_per_node = self.tasks_per_node(queue_id)
        num_nodes = int(math.ceil(float(num_tasks) / float(tasks_per_node)))

        if threads_per_task is None:
            threads_per_task = 1

        if ":" not in wall_clock:
            wall_clock = wall_clock + ":00"
        elif len(wall_clock.split(":")) == 3:
            # LSF format hh:mm, remove seconds
            wall_clock = wall_clock.rsplit(":", 1)[0]

        if "mycluster-" in job_script:
            my_script = self._get_data(my_script)

        num_queue_slots = num_nodes * self.tasks_per_node(queue_id)

        if output_name is None:
            output_name = job_name + ".out"

        template = self._load_template("lsf.jinja")

        script_str = template.render(
            my_name=job_name,
            my_script=job_script,
            my_output=output_name,
            user_email=user_email,
            queue_name=queue_id,
            num_queue_slots=num_queue_slots,
            num_tasks=num_tasks,
            tpn=tasks_per_node,
            num_threads_per_task=threads_per_task,
            num_nodes=num_nodes,
            project_name=project_name,
            wall_clock=wall_clock,
            openmpi_args=openmpi_args,
            qos=qos,
            exclusive=exclusive,
        )

        return script_str

    def submit(
        self, script_name, immediate=False, depends_on=None, depends_on_always_run=False
    ):
        job_id = None
        if immediate:
            raise NotYetImplementedException("Immediate not yet implemented for LSF")
        if depends_on and depends_on_always_run:
            cmd = 'bsub -w "ended(%s)" < %s ' % (depends_on, script_name)
            with os.popen(cmd) as f:
                output = f.readline()
                try:
                    job_id = int(output.split(" ")[1].replace("<", "").replace(">", ""))
                except:
                    raise SchedulerException("Job submission failed: " + output)
        elif depends_on is not None:
            cmd = 'bsub -w "done(%s)" < %s ' % (depends_on, script_name)
            with os.popen(cmd) as f:
                output = f.readline()
                try:
                    job_id = int(output.split(" ")[1].replace("<", "").replace(">", ""))
                except:
                    raise SchedulerException("Job submission failed: " + output)
        else:
            with os.popen("bsub <" + script_name) as f:
                output = f.readline()
                try:
                    job_id = int(output.split(" ")[1].replace("<", "").replace(">", ""))
                except:
                    raise SchedulerException("Job submission failed: " + output)
        return job_id

    def list_current_jobs(self):
        jobs = []
        output = subprocess.run(
            'bjobs -noheader -u `whoami` -o "jobid queue job_name stat"',
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
        )
        if output.returncode == 0:
            for line in output.stdout.decode("utf-8").splitlines():
                if line == "No unfinished job found":
                    return jobs
                job_info = re.sub(" +", " ", line.strip()).split(" ")
                jobs.append(
                    {
                        "id": int(job_info[0]),
                        "queue": job_info[1],
                        "name": job_info[2],
                        "state": job_info[3],
                    }
                )
        else:
            raise SchedulerException("Error fetching job queue listing")
        return jobs

    def get_job_details(self, job_id):
        """
        Get full job and step stats for job_id
        """
        stats_dict = {}
        with os.popen(
            "bjobs -o \"jobid run_time cpu_used  queue slots  stat exit_code start_time estimated_start_time finish_time delimiter='|'\" -noheader "
            + str(job_id)
        ) as f:
            try:
                line = f.readline()
                cols = line.split("|")
                stats_dict["job_id"] = cols[0]
                if cols[1] != "-":
                    stats_dict["wallclock"] = datetime.timedelta(
                        seconds=float(cols[1].split(" ")[0])
                    )
                if cols[2] != "-":
                    stats_dict["cpu"] = datetime.timedelta(seconds=float(cols[2].split(" ")[0]))
                stats_dict["queue"] = cols[3]
                stats_dict["status"] = cols[5]
                stats_dict["exit_code"] = cols[6]
                stats_dict["start"] = cols[7]
                stats_dict["start_time"] = cols[8]
                if stats_dict["status"] in ["DONE", "EXIT"]:
                    stats_dict["end"] = cols[9]
            except:
                with os.popen("bhist -l " + str(job_id)) as f:
                    try:
                        output = f.readlines()
                        for line in output:
                            if "Done successfully" in line:
                                stats_dict["status"] = "DONE"
                                return stats_dict
                            elif "Completed <exit>" in line:
                                stats_dict["status"] = "EXIT"
                                return stats_dict
                            else:
                                stats_dict["status"] = "UNKNOWN"
                    except Exception as e:
                        raise SchedulerException(
                            "Error fetching job details from bjobs, check job id."
                        )
        return stats_dict

    def delete(self, job_id):
        cmd = f"bkill {job_id}"
        output = subprocess.run(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
        )
        if output.returncode != 0:
            raise SchedulerException(f"Error cancelling job {job_id}")



# BSD 3 - Clause License

# Copyright(c) 2021, Zenotech
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and / or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#         SERVICES
#         LOSS OF USE, DATA, OR PROFITS
#         OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .slurm import Slurm



# BSD 3 - Clause License

# Copyright(c) 2021, Zenotech
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and / or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#         SERVICES
#         LOSS OF USE, DATA, OR PROFITS
#         OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import errno
import click
import pyfiglet
from configparser import ConfigParser

import mycluster
from .exceptions import ConfigurationException, SchedulerException

APP_NAME = "MyCluster"


def read_config():
    cfg = os.path.join(click.get_app_dir(APP_NAME), "config.ini")
    if os.path.isfile(cfg):
        parser = ConfigParser(allow_no_value=True)
        parser.read([cfg])
        email = parser.get("default", "email")
        return {"email": email, "silent": True}
    return {"email": "None", "silent": True}


@click.group()
@click.pass_context
@click.option("-s", "--silent", is_flag=True, help="Hide application banner")
@click.option(
    "--email",
    type=str,
    help="Email address to use in submission files",
)
def main(ctx, silent, email):
    """CLI for MyCluster"""
    config = read_config()
    if email is not None:
        config["email"] = email
    if not silent:
        config["silent"] = False
        click.echo(pyfiglet.Figlet().renderText("MyCluster"))
        click.echo(f"User email: '{config['email']}'")
    try:
        scheduler = mycluster.detect_scheduling_sys()
        if not silent:
            click.echo(f"Scheduler '{scheduler.scheduler_type()}' detected")
        ctx.obj = (scheduler, config)
    except ConfigurationException as e:
        click.echo(e)
        exit(1)


@main.command()
@click.pass_context
def configure(ctx):
    """Configure MyCluster to save your email address"""
    email = click.prompt("Please enter your email address")
    config = ConfigParser()
    if not config.has_section("default"):
        config.add_section("default")
    config.set("default", "email", email)
    cfg = os.path.join(click.get_app_dir(APP_NAME), "config.ini")
    try:
        os.makedirs(os.path.dirname(cfg))
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
    with open(cfg, "w") as configfile:
        config.write(configfile)
        click.echo(f"Config file written to {cfg}")


@main.command()
@click.pass_context
def queues(ctx):
    """List all available queues"""
    click.echo(
        "{0:25} | {1:^15} | {2:^15} | {3:^15} | {4:^15} | {5:^15}".format(
            "Queue Name",
            "Node Max Task",
            "Node Max Thread",
            "Node Max Memory",
            "Max Task",
            "Available Task",
        )
    )
    for q in ctx.obj[0].queues():
        try:
            nc = ctx.obj[0].node_config(q)
            tpn = ctx.obj[0].tasks_per_node(q)
            avail = ctx.obj[0].available_tasks(q)
        except SchedulerException:
            nc = None
            tpn = None
            avail = None
        click.echo(
            f'{q:25} | {tpn:^15} | {nc["max thread"]:^15} | {nc["max memory"]:^15} | {nc["max task"]:^15} | {avail["available"]:^15}'
        )


@click.argument("runscript")
@click.argument("queue")
@click.argument("jobfile")
@click.option(
    "-j",
    "--jobname",
    default="myclusterjob",
    help="Name to give job",
    show_default=True,
)
@click.option("-n", "--ntasks", default=1, help="Number of tasks", show_default=True)
@click.option(
    "-tpt",
    "--threadspertask",
    type=int,
    required=False,
    help="Number of threads per task",
)
@click.option(
    "-tpn",
    "--taskpernode",
    type=int,
    required=False,
    help="Task per node",
)
@click.option(
    "-p",
    "--project",
    default="default",
    help="Project/Account used of accounting usage",
    show_default=True,
)
@click.option(
    "-q",
    "--qos",
    default="default",
    help="QOS level to submit to",
    show_default=True,
)
@click.option(
    "--ompiargs",
    default="-bysocket -bind-to-socket",
    help="Arguments for OpenMPI",
    show_default=True,
)
@click.option(
    "--maxtime",
    default="12:00:00",
    help="Maximum runtime (hh:mm::ss)",
    show_default=True,
)
@click.option(
    "--shared",
    is_flag=True,
    help="Request shared node allocation",
)
@main.command()
@click.pass_context
def create(
    ctx,
    runscript,
    queue,
    jobfile,
    jobname,
    ntasks,
    threadspertask,
    taskpernode,
    project,
    qos,
    ompiargs,
    maxtime,
    shared,
):
    """Create a job file to submit RUNSCRIPT to QUEUE and write it to JOBFILE"""
    if "email" in ctx.obj[1]:
        user_email = ctx.obj[1]["email"]
    else:
        user_email = None
    script = ctx.obj[0].create_submit(
        queue,
        ntasks,
        jobname,
        runscript,
        maxtime,
        ompiargs,
        project,
        tasks_per_node=taskpernode,
        threads_per_task=threadspertask,
        user_email=user_email,
        qos=qos,
        exclusive=not shared,
    )
    with click.open_file(jobfile, "w") as f:
        f.write(script)
    click.echo(f"Job file written to {jobfile}")


@click.argument("jobfile")
@click.option(
    "--immediate",
    is_flag=True,
    help="Submit job immediately",
)
@click.option(
    "--depends",
    type=str,
    help="List of job dependencies jobA_id:jobB_id:jobC_id (Slurm only)",
)
@main.command()
@click.pass_context
def submit(ctx, jobfile, immediate, depends):
    """Submit a batch script JOBFILE to the scheduler"""
    if os.path.isfile(jobfile):
        job_id = ctx.obj[0].submit(jobfile, immediate, depends)
        click.echo(f"Job submitted with ID '{job_id}'")
    else:
        click.echo(f"Error: Job file '{jobfile}' not found.")


@main.command()
@click.pass_context
def list(ctx):
    """List the status of active jobs"""
    print(
        "{0:^10} | {1:^10} | {2:^10} | {3:^12}".format(
            "Job ID", "Job Name", "Queue", "Status"
        )
    )
    for job in ctx.obj[0].list_current_jobs():
        print(
            f"{job['id']:^10} | {job['name']:^10} | {job['queue']:^10} | {job['state']:^12}"
        )


@click.argument("jobid", type=int)
@main.command()
@click.pass_context
def details(ctx, jobid):
    """Get the details of job with id JOBID"""
    try:
        click.echo(ctx.obj[0].get_job_details(jobid))
    except SchedulerException as e:
        click.echo(e)


@click.argument("jobid", type=int)
@main.command()
@click.pass_context
def cancel(ctx, jobid):
    """Cancel job with id JOBID"""
    try:
        click.echo(ctx.obj[0].delete(jobid))
    except SchedulerException as e:
        click.echo(e)


if __name__ == "__main__":
    main()



# BSD 3 - Clause License

# Copyright(c) 2021, Zenotech
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and / or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#         SERVICES
#         LOSS OF USE, DATA, OR PROFITS
#         OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import sys
import logging
from abc import ABC, abstractmethod
from jinja2 import Environment, FileSystemLoader
from datetime import timedelta
from subprocess import check_output

logger = logging.getLogger(__name__)


class Scheduler(ABC):
    """
    Abstract base class for a batch scheduler interface
    """

    @abstractmethod
    def scheduler_type(self):
        """
        Returns the scheduler type
        """
        pass

    @abstractmethod
    def name(self):
        """
        Fetch the cluster name
        """
        pass

    @abstractmethod
    def queues(self):
        """
        List the available queues
        """
        pass

    @abstractmethod
    def node_config(self, queue_id):
        """
        Fetch the hardware config for nodes in queue with queue_id
        """
        pass

    @abstractmethod
    def tasks_per_node(self, queue_id):
        """
        How many tasks can be run per node in queue with queue_id
        """
        pass

    @abstractmethod
    def available_tasks(self, queue_id):
        """
        Get the current status of the cluster
        """
        pass

    @abstractmethod
    def create_submit(
        self,
        queue_id,
        num_tasks,
        job_name,
        job_script,
        wall_clock,
        openmpi_args="-bysocket -bind-to-socket",
        project_name="default",
        tasks_per_node=None,
        threads_per_task=1,
        user_email=None,
        qos=None,
        exclusive=True,
        output_name=None,
        gres=None,
    ):
        """
        Write a new job file
        """
        pass

    @abstractmethod
    def submit(
        self, script_name, immediate=False, depends_on=None, depends_on_always_run=False
    ):
        """
        Submit the job file script_name to the scheduler
        """
        pass

    @abstractmethod
    def list_current_jobs(self):
        """
        List all the jobs currently active with the scheduler
        """
        pass

    @abstractmethod
    def get_job_details(self, job_id):
        """
        Get the details of job with ID job_id
        """
        pass

    @abstractmethod
    def delete(self, job_id):
        """
        Delete job with ID job_id
        """
        pass

    def accounts(self):
        """
        List billings accounts configured in scheduler
        """
        return []

    def _get_template_name(self):
        if "MYCLUSTER_TEMPLATE" in os.environ:
            logger.debug(
                f'MYCLUSTER_TEMPLATE set, loading template from {os.environ["MYCLUSTER_TEMPLATE"]}'
            )
            return os.environ["MYCLUSTER_TEMPLATE"]
        else:
            return f"{self.scheduler_type()}.jinja"

    def _load_template(self, template_name):
        """
        Load the jinja2 template with name template_name
        1. Check if template_name is absolute
        2. Try loading it from templates dir
        """
        if os.path.isfile(template_name):
            logger.debug(f'File "{template_name}" exists, loading it.')
            env = Environment(loader=FileSystemLoader(os.path.dirname(template_name)))
            return env.get_template(os.path.basename(template_name))
        else:
            logger.debug(
                f'Loading template "{template_name}" from "{os.path.join(os.path.dirname(__file__), "templates")}"'
            )
            env = Environment(
                loader=FileSystemLoader(
                    os.path.join(os.path.dirname(__file__), "templates")
                )
            )
            return env.get_template(template_name)

    def _get_data(self, filename):
        """
        Helper function to find the application specific run-scripts
        """
        packagedir = os.path.dirname(__file__)
        # Need to traverse lib/python3.x/site-packages/mycluster
        # So 4 directories above the current packagedir
        dirname = os.path.join(packagedir, "..", "..", "..", "..", "share", "MyCluster")
        fullname = os.path.join(dirname, filename)

        if not os.path.isfile(fullname):
            dirname = os.path.join(packagedir, "..", "share", "MyCluster")
            fullname = os.path.join(dirname, filename)
        # Need to check if file exists as
        # share location may also be sys.prefix/share
        if not os.path.isfile(fullname):
            dirname = os.path.join(sys.prefix, "share", "MyCluster")
            fullname = os.path.join(dirname, filename)

        if not os.path.isfile(fullname):
            raise Exception(f"Unable to find file {fullname}")

        logger.debug(f'Loading data file from "{fullname}"')
        return fullname

    def _get_timedelta(self, date_str):
        """
        Converts date_str from string in [DD-[hh:]]mm:ss format to a timedelta object
        """
        days = 0
        hours = 0
        minutes = 0
        seconds = 0

        if date_str.count("-") == 1:
            days = int(date_str.split("-")[0])
            date_str = date_str.partition("-")[2]
        if date_str.count(":") == 2:
            hours = int(date_str.split(":")[0])
            date_str = date_str.partition(":")[2]

        try:
            minutes = int(date_str.split(":")[0])
            seconds = int(date_str.split(":")[1])
        except:
            pass

        return timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds)

    def _check_output(self, command_list):
        """
        check_output wrapper that decodes to a str instead of bytes
        """
        return check_output(command_list).decode("UTF-8")



# BSD 3 - Clause License

# Copyright(c) 2021, Zenotech
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and / or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#         SERVICES
#         LOSS OF USE, DATA, OR PROFITS
#         OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
import re
import math
import subprocess

from .base import Scheduler
from mycluster.exceptions import SchedulerException, ConfigurationException


class Slurm(Scheduler):
    def scheduler_type(self):
        return "slurm"

    def name(self):
        with os.popen("sacctmgr show cluster") as f:
            f.readline()
            f.readline()
            return f.readline().strip().split(" ")[0]

    def queues(self):
        queue_list = []
        with os.popen("sinfo -sh") as f:
            for line in f:
                q = line.split(" ")[0].strip().replace("*", "")
                queue_list.append(q)
        return queue_list

    def node_config(self, queue_id):
        # Find first node with queue and record node config
        queue_name = queue_id
        tasks = 0
        config = {}
        with os.popen("sinfo -Nelh -p " + queue_name) as f:
            line = f.readline()
            if len(line):
                new_line = re.sub(" +", " ", line.strip())
                tasks = int(new_line.split(" ")[4])
                memory = int(new_line.split(" ")[6])
                config["max task"] = tasks
                config["max thread"] = tasks
                config["max memory"] = memory
            else:
                raise SchedulerException(
                    "Requested partition %s has no nodes" % queue_name
                )
        return config

    def tasks_per_node(self, queue_id):
        queue_name = queue_id
        tasks = 0
        with os.popen("sinfo -Nelh -p " + queue_name) as f:
            line = f.readline()
            new_line = re.sub(" +", " ", line.strip())
            tasks = int(new_line.split(" ")[4])
        return tasks

    def available_tasks(self, queue_id):
        free_tasks = 0
        max_tasks = 0
        queue_name = queue_id
        nc = self.node_config(queue_id)
        with os.popen("sinfo -sh -p " + queue_name) as f:
            line = f.readline()
            new_line = re.sub(" +", " ", line.strip())
            line = new_line.split(" ")[3]
            free_tasks = int(line.split("/")[1]) * nc["max task"]
            max_tasks = int(line.split("/")[3]) * nc["max task"]
        return {"available": free_tasks, "max tasks": max_tasks}

    def accounts(self):
        account_list = []
        with os.popen(
            "sacctmgr --noheader list assoc user=`id -un` format=Account"
        ) as f:
            for line in f:
                account_list.append(line)
        return account_list

    def create_submit(
        self,
        queue_id,
        num_tasks,
        job_name,
        job_script,
        wall_clock,
        openmpi_args="-bysocket -bind-to-socket",
        project_name="default",
        tasks_per_node=None,
        threads_per_task=1,
        user_email=None,
        qos=None,
        exclusive=True,
        output_name=None,
        gres=None
    ):
        queue_name = queue_id
        if tasks_per_node is None:
            tasks_per_node = self.tasks_per_node(queue_id)
        num_nodes = int(math.ceil(float(num_tasks) / float(tasks_per_node)))
        if threads_per_task is None:
            threads_per_task = -1

        if ":" not in wall_clock:
            wall_clock = wall_clock + ":00:00"

        if "mycluster-" in job_script:
            job_script = self._get_data(job_script)

        if output_name is None:
            output_name = job_name + ".out"

        template = self._load_template(self._get_template_name())

        script_str = template.render(
            my_name=job_name,
            my_script=job_script,
            my_output=output_name,
            user_email=user_email,
            queue_name=queue_name,
            num_tasks=num_tasks,
            tpn=tasks_per_node,
            num_threads_per_task=threads_per_task,
            num_nodes=num_nodes,
            project_name=project_name,
            wall_clock=wall_clock,
            openmpi_args=openmpi_args,
            qos=qos,
            exclusive=exclusive,
            gres=gres
        )
        return script_str

    def submit(
        self, script_name, immediate=False, depends_on=None, depends_on_always_run=False
    ):
        additional_cmd = ""
        if "MYCLUSTER_SUBMIT_OPT" in os.environ:
            additional_cmd = os.environ["MYCLUSTER_SUBMIT_OPT"]
        if not immediate:
            if depends_on and depends_on_always_run:
                command = f"sbatch {additional_cmd} --kill-on-invalid-dep=yes --dependency=afterany:{depends_on} {script_name}"
            elif depends_on is not None:
                command = f"sbatch {additional_cmd} --kill-on-invalid-dep=yes --dependency=afterok:{depends_on} {script_name}"
            else:
                command = f"sbatch {additional_cmd} {script_name}"
            print(f"running {command}")
            with os.popen(command) as f:
                output = f.readline()
                try:
                    job_id = int(output.split(" ")[-1].strip())
                except:
                    raise SchedulerException("Job submission failed: " + output)
        else:
            with os.popen(
                'grep -- "SBATCH -p" ' + script_name + " | sed 's/#SBATCH//'"
            ) as f:
                partition = f.readline().rstrip()
            with os.popen(
                'grep -- "SBATCH --nodes" ' + script_name + " | sed 's/#SBATCH//'"
            ) as f:
                nnodes = f.readline().rstrip()
            with os.popen(
                'grep -- "SBATCH --ntasks" ' + script_name + " | sed 's/#SBATCH//'"
            ) as f:
                ntasks = f.readline().rstrip()
            with os.popen(
                'grep -- "SBATCH -A" ' + script_name + " | sed 's/#SBATCH//'"
            ) as f:
                project = f.readline().rstrip()
            with os.popen(
                'grep -- "SBATCH -J" ' + script_name + " | sed 's/#SBATCH//'"
            ) as f:
                job = f.readline().rstrip()

            cmd_line = (
                "salloc --exclusive "
                + nnodes
                + " "
                + partition
                + " "
                + ntasks
                + " "
                + project
                + " "
                + job
                + " bash ./"
                + script_name
            )
            try:
                output = subprocess.check_output(cmd_line, shell=True)
                try:
                    job_id = int(output.split(" ")[-1].strip())
                except:
                    raise SchedulerException("Job submission failed: " + output)
            except:
                raise SchedulerException("Job submission failed: " + cmd_line)
        return job_id

    def list_current_jobs(self):
        jobs = []
        output = subprocess.run(
            "squeue -h -u `whoami`",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
        )
        if output.returncode == 0:
            for line in output.stdout.decode("utf-8").splitlines():
                job_info = re.sub(" +", " ", line.strip()).split(" ")
                jobs.append(
                    {
                        "id": int(job_info[0]),
                        "queue": job_info[1],
                        "name": job_info[2],
                        "state": job_info[4],
                    }
                )
        else:
            raise SchedulerException("Error fetching job queue listing")
        return jobs

    def get_job_details(self, job_id):
        """
        Get full job and step stats for job_id
        First check using sacct, then fallback to squeue
        """
        stats_dict = {}
        sacct_cmd = f"sacct --noheader --format JobId,Elapsed,TotalCPU,Partition,NTasks,AveRSS,State,ExitCode,start,end -P -j {job_id}"
        squeue_cmd = f'squeue --format "%.18i %.9P %.8j %.8u %.2t %.10M %.6D %R %S" -h -j {job_id}'
        output = subprocess.run(
            sacct_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
        )
        if output.returncode != 0:
            raise SchedulerException("Error fetching job details from sacct")
        lines = output.stdout.decode("utf-8").splitlines()
        if len(lines) != 0:
            if lines[0] not in [
                "SLURM accounting storage is disabled",
                "slurm_load_jobs error: Invalid job id specified",
            ]:
                cols = lines[0].split("|")
                stats_dict["job_id"] = cols[0]
                stats_dict["wallclock"] = self._get_timedelta(cols[1])
                stats_dict["cpu"] = self._get_timedelta(cols[2])
                stats_dict["queue"] = cols[3]
                stats_dict["status"] = cols[6]
                stats_dict["exit_code"] = cols[7].split(":")[0]
                stats_dict["start"] = cols[8]
                stats_dict["end"] = cols[9]
                steps = []
                for line in lines[1:]:
                    step = {}
                    cols = line.split("|")
                    step_val = cols[0].split(".")[1]
                    step["step"] = step_val
                    step["wallclock"] = self._get_timedelta(cols[1])
                    step["cpu"] = self._get_timedelta(cols[2])
                    step["ntasks"] = cols[4]
                    step["status"] = cols[6]
                    step["exit_code"] = cols[7].split(":")[0]
                    step["start"] = cols[8]
                    step["end"] = cols[9]
                    steps.append(step)
                stats_dict["steps"] = steps
        else:
            output = subprocess.run(
                squeue_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
            )
            if output.returncode != 0:
                raise SchedulerException(
                    "Error fetching job details from squeue, check job id."
                )
            lines = output.stdout.decode("utf-8").splitlines()
            for line in lines:
                if line == "slurm_load_jobs error: Invalid job id specified":
                    raise SchedulerException("Invalid job id specified")
                new_line = re.sub(" +", " ", line.strip())
                job_id = int(new_line.split(" ")[0])
                state = new_line.split(" ")[4]
                stats_dict["job_id"] = str(job_id)
                stats_dict["status"] = state
                stats_dict["start_time"] = new_line.split(" ")[8]
        return stats_dict

    def delete(self, job_id):
        cmd = f"scancel {job_id}"
        output = subprocess.run(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
        )
        if output.returncode != 0:
            raise SchedulerException(f"Error cancelling job {job_id}")



# BSD 3 - Clause License

# Copyright(c) 2021, Zenotech
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and / or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#         SERVICES
#         LOSS OF USE, DATA, OR PROFITS
#         OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
import re
import math
import subprocess

from .base import Scheduler
from mycluster.exceptions import (
    SchedulerException,
    ConfigurationException,
    NotYetImplementedException,
)


class PBS(Scheduler):
    def scheduler_type(self):
        return "pbs"

    def name(self):
        raise NotYetImplementedException("name() not implemented for PBS")

    def queues(self):
        queue_list = []
        try:
            output = self._check_output(["qstat", "-Q"])
            lines = output.splitlines()[2:]
            for queue in lines:
                queue_list.append(queue.split(" ")[0])
        except Exception as e:
            raise SchedulerException("Error fetching queues")
        return queue_list

    def _get_vnode_name(self, queue_id):
        try:
            output = self._check_output(["qstat", "-Qf", queue_id])
            for line in output.splitlines():
                if line.strip().startswith("default_chunk.vntype"):
                    return line.split("=")[-1].strip()
        except Exception as e:
            raise SchedulerException("Error fetching node config")

    def node_config(self, queue_id):
        max_threads = 1
        max_memory = 1
        try:
            tpn = tasks_per_node(queue_id)
            vnode_type = vnode_type = vnode_type = self._get_vnode_name(queue_id)
            if vnode_type is not None:
                output = self._check_output(
                    "pbsnodes -a -F dsv | grep {}".format(vnode_type), shell=True
                )
            for line in output.splitlines():
                for item in line.split("|"):
                    [key, value] = item.strip().split("=")
                    if key.strip() == "resources_available.vps_per_ppu":
                        if int(value) > max_threads:
                            max_threads = int(value) * tpn
                    if key.strip() == "resources_available.mem":
                        # strip kb and convert to mb
                        mem = float(value[:-2]) / 1024
                        if mem > max_memory:
                            max_memory = mem
        except Exception as e:
            raise SchedulerException("Error fetching node config")
        return {"max thread": max_threads, "max memory": max_memory}

    def tasks_per_node(self, queue_id):
        tpn = 1
        try:
            vnode_type = vnode_type = self._get_vnode_name(queue_id)
            if vnode_type is not None:
                output = self._check_output(
                    "pbsnodes -a -F dsv | grep {}".format(vnode_type), shell=True
                )
                for line in output.splitlines():
                    for item in line.split("|"):
                        [key, value] = item.strip().split("=")
                        if key.strip() == "resources_available.ncpus":
                            if int(value) > tpn:
                                tpn = int(value)
        except Exception as e:
            raise SchedulerException("Error fetching node config")
        return tpn

    def available_tasks(self, queue_id):
        free_tasks = 0
        max_tasks = 0
        assigned_tasks = 0
        try:
            vnode_type = self._get_vnode_name(queue_id)
            if vnode_type is not None:
                output = self._check_output(
                    "pbsnodes -a -F dsv | grep {}".format(vnode_type), shell=True
                )
            for line in output.splitlines():
                for item in line.split("|"):
                    [key, value] = item.strip().split("=")
                    if key.strip() == "resources_available.ncpus":
                        max_tasks += int(value)
                    elif key.strip() == "resources_assigned.ncpus":
                        assigned_tasks += int(value)
            free_tasks = max_tasks - assigned_tasks
        except Exception as e:
            raise SchedulerException("Error fetching node config")
        return {"available": free_tasks, "max tasks": max_tasks}

    def _min_tasks_per_node(self, queue_id):
        return 1

    def create_submit(
        self,
        queue_id,
        num_tasks,
        job_name,
        job_script,
        wall_clock,
        openmpi_args="-bysocket -bind-to-socket",
        project_name="default",
        tasks_per_node=None,
        threads_per_task=1,
        user_email=None,
        qos=None,
        exclusive=True,
        output_name=None,
        gres=None
    ):
        queue_name = queue_id
        if tasks_per_node is None:
            tasks_per_node = self.tasks_per_node(queue_id)
        num_nodes = int(math.ceil(float(num_tasks) / float(tasks_per_node)))

        if threads_per_task is None:
            threads_per_task = 1

        if "mycluster-" in my_script:
            my_script = self._get_data(my_script)

        if ":" not in wall_clock:
            wall_clock = wall_clock + ":00:00"

        num_nodes = int(math.ceil(float(num_tasks) / float(tasks_per_node)))

        num_queue_slots = num_nodes * queue_tpn

        if not exclusive:
            if num_nodes == 1:  # Assumes fill up rule
                num_queue_slots = num_nodes * max(
                    tasks_per_node, self._min_tasks_per_node(queue_id)
                )

        if output_name is None:
            output_name = job_name + ".out"

        template = self._load_template("pbs.jinja")

        script_str = template.render(
            my_name=job_name,
            my_script=job_script,
            my_output=output_name,
            user_email=user_email,
            queue_name=queue_name,
            num_queue_slots=num_queue_slots,
            num_tasks=num_tasks,
            tpn=tasks_per_node,
            num_threads_per_task=threads_per_task,
            num_nodes=num_nodes,
            project_name=project_name,
            wall_clock=wall_clock,
            openmpi_args=openmpi_args,
            qos=qos,
        )

        return script_str

    def submit(
        self, script_name, immediate=False, depends_on=None, depends_on_always_run=False
    ):
        job_id = None
        if not immediate:
            if depends_on and depends_on_always_run:
                with os.popen(
                    "qsub -W depend=afterany:%s %s" % (depends_on, script_name)
                ) as f:
                    output = f.readline()
                    try:
                        job_id = output.strip().split(".")[0]
                    except:
                        raise SchedulerException("Job submission failed: " + output)
            elif depends_on is not None:
                with os.popen(
                    "qsub -W depend=afterok:%s %s" % (depends_on, script_name)
                ) as f:
                    output = f.readline()
                    try:
                        job_id = output.strip().split(".")[0]
                    except:
                        raise SchedulerException("Job submission failed: " + output)
            else:
                with os.popen("qsub " + script_name) as f:
                    output = f.readline()
                    try:
                        job_id = output.strip().split(".")[0]
                    except:
                        raise SchedulerException("Job submission failed: " + output)
        else:
            raise NotYetImplementedException("Immediate not yet implemented for PBS")
        return job_id

    def list_current_jobs(self):
        raise NotYetImplementedException(
            "list_current_jobs not yet implemented for PBS"
        )

    def get_job_details(self, job_id):
        """
        Get full job and step stats for job_id
        """
        stats_dict = {}
        with os.popen("qstat -xf " + str(job_id)) as f:
            try:
                line = f.readline().strip()
                while line:
                    if line.startswith("Job Id:"):
                        stats_dict["job_id"] = line.split(":")[1].split(".")[0].strip()
                    elif line.startswith("resources_used.walltime"):
                        stats_dict["wallclock"] = get_timedelta(line.split("=")[1])
                    elif line.startswith("resources_used.cput"):
                        stats_dict["cpu"] = get_timedelta(line.split("=")[1])
                    elif line.startswith("queue"):
                        stats_dict["queue"] = line.split("=")[1].strip()
                    elif line.startswith("job_state"):
                        stats_dict["status"] = line.split("=")[1].strip()
                    elif line.startswith("Exit_status"):
                        stats_dict["exit_code"] = line.split("=")[1].strip()
                    elif line.startswith("Exit_status"):
                        stats_dict["exit_code"] = line.split("=")[1].strip()
                    elif line.startswith("stime"):
                        stats_dict["start"] = line.split("=")[1].strip()
                    line = f.readline().strip()
                if stats_dict["status"] == "F" and "exit_code" not in stats_dict:
                    stats_dict["status"] = "CA"
                elif stats_dict["status"] == "F" and stats_dict["exit_code"] == "0":
                    stats_dict["status"] = "PBS_F"
            except Exception as e:
                with os.popen("qstat -xaw " + str(job_id)) as f:
                    try:
                        output = f.readlines()
                        for line in output:
                            if str(job_id) in line:
                                cols = line.split()
                                stats_dict["job_id"] = cols[0].split(".")[0]
                                stats_dict["queue"] = cols[2]
                                stats_dict["status"] = cols[9]
                                stats_dict["wallclock"] = get_timedelta(cols[10])
                                return stats_dict
                    except Exception as e:
                        stats_dict["status"] = "UNKNOWN"
        return stats_dict

    def delete(self, job_id):
        cmd = f"qdel {job_id}"
        output = subprocess.run(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
        )
        if output.returncode != 0:
            raise SchedulerException(f"Error cancelling job {job_id}")
